
#include <XSContainer.h>
#include <XSstreams.h>
#include <XSFit/Fit/Fit.h>
#include <XSModel/Data/SpectralData.h>
#include <XSModel/Data/BackCorr/Background.h>
#include <XSModel/Data/Detector/Response.h>
#include <XSModel/GlobalContainer/DataContainer.h>
#include <XSModel/GlobalContainer/ModelContainer.h>
#include <XSModel/Model/Model.h>
#include <XSModel/Model/Component/SumComponent.h>
#include <XSPlot/Plot/PlotGroup.h>
#include <XSPlot/Plot/PlotSettings.h>
#include <XSPlot/Plot/PlotVector.h>
#include <XSStat/ChiSquare.h>
#include <XSUtil/Error/Error.h>
#include <XSUtil/Numerics/Numerics.h>
#include <algorithm>
#include <sstream>
#include <utility>

// CreatePolarizationPlotGroups
#include <XSPlot/Plot/CreatePolarizationPlotGroups.h>


// Class CreatePolarizationPlotGroups 

CreatePolarizationPlotGroups::CreatePolarizationPlotGroups (bool isAngle)
  : PlotGroupCreator(),
   m_isAngle(isAngle),
   m_modelPositionDirectory(),
   m_modelsForSpectra(),
   m_settings(0) // Non-owning
{
}


CreatePolarizationPlotGroups::~CreatePolarizationPlotGroups()
{
}


std::vector<PlotGroup*> CreatePolarizationPlotGroups::createPlotGroups (const PlotSettings& settings, const IntegerVector& plotGroupNumbers)
{
  using namespace XSContainer;   
  m_settings = &settings;
  std::vector<PlotGroup*> plotGroups;

  // First construct a vector of spectra to include in each PlotGroup. All spectra in a datagroup are assigned
  // to the same PlotGroup. If any spectra in another datagroup have the same plotGroupNumber as a spectrum
  // in the current datagroup then all spectra in this datagroup are also assigned to the PlotGroup.

  std::vector<std::vector<SpectralData*> > groupedSpectra = divideIntoGroups();

  // groupedSpectra now contains a vector of pointers to spectra for each PlotGroup that we want to create

  itemizePlotGroupModels(groupedSpectra);

  // This function owns the new PlotGroups until it returns.
  // Therefore it's responsible for cleanup if anything throws.
  try
  {
     // loop over the PlotGroups to be created
     for (size_t j = 0; j < groupedSpectra.size(); ++j )
     {
        int chanXGroupOffset (0); // icbeg
        // can throw an "EnergiesNotDefined" exception...
        // Effective areas will be initialized in here (if needed).
        PlotGroup* gr = initializePlotGroup((j+1),groupedSpectra[j]);
        plotGroups.push_back(gr);

        // gr->m_n can be 0 for the case of ALL channels ignored.  Otherwise
        // n does NOT take ignored channels into account.
        if (gr->m_n > 0)
        {
           int npts (0);
           int latestChannel(0);
           bool moreBins (true);
           do // execute at least once...
           {
	     if ( (moreBins = getNextPlotBin(gr, npts, groupedSpectra[j],
					     m_modelPositionDirectory[j], latestChannel)) )
              {
                 if (m_settings->xOption() == CHANNELS) 
                     gr->m_xAxis.data[npts] += chanXGroupOffset;
                 ++npts;
              }
	      else gr->m_n = npts;

           }  while ( moreBins );
	   clearEffectiveAreas(groupedSpectra[j]);
           if (gr->m_model.size() > 1)
           {
	      std::vector<Real>& modelSum = gr->m_model[0].data;
              for (size_t l = 1; l < gr->m_model.size(); ++l)
              {
	         std::vector<Real>& model = gr->m_model[l].data;
                 for (int k = 0; k < npts; ++k) 
                 {
                    modelSum[k] += model[k];       
                 }
              }
           }
           chanXGroupOffset += gr->m_channelPlotSpacing;
        } // end if gr->m_n > 0 
     } // end plot groups loop
  }
  catch (YellowAlert&)
  {
     for (size_t i=0; i<plotGroups.size(); ++i)
     {
        // Whether or not initEffectiveAreas was called, or
        // effectiveAreas have already been cleared for the
        // particular group, this is safe to do:
        clearEffectiveAreas(groupedSpectra[i]);
        delete plotGroups[i];
     }
     throw;
  }

  // if we are creating a polarization angle plot then by construction all the
  // angles lie between 0 and 180 degrees. If the data values lie close to 0 or
  // 180 then the plotted values could jump between the bottom and top of the plot
  // in successive bins which would be confusing. To avoid this we calculate the
  // average angle of the data and then wrap all angles onto this average +/- 90 degrees.

  if ( m_isAngle && plotGroups.size() > 0 ) {

    Real meanAngle(0.0);
    for (size_t igr=0; igr<plotGroups.size(); igr++) {
      meanAngle += plotGroups[igr]->getDataMean();
    }
    meanAngle /= plotGroups.size();
    Real minAngle = meanAngle - 90.0;
    Real maxAngle = meanAngle + 90.0;

    for (size_t igr=0; igr<plotGroups.size(); igr++) {
      PlotGroup* gr = plotGroups[igr];
      for (size_t ibin=0; ibin<gr->m_n; ibin++) {
	if ( gr->m_yData.data[ibin] < minAngle ) gr->m_yData.data[ibin] += 180.0;
	if ( gr->m_yData.data[ibin] > maxAngle ) gr->m_yData.data[ibin] -= 180.0;
	if ( gr->m_model.size() > 0 ) {
	  if ( gr->m_model[0].data[ibin] < minAngle ) gr->m_model[0].data[ibin] += 180.0;
	  if ( gr->m_model[0].data[ibin] > maxAngle ) gr->m_model[0].data[ibin] -= 180.0;
	}
      }
    }

  }

  return plotGroups;
}

void CreatePolarizationPlotGroups::itemizePlotGroupModels (const std::vector<std::vector<SpectralData*> >& groupedSpectra)
{
   using namespace XSContainer;
   // Only count active/on models.
   m_modelPositionDirectory.clear();
   m_modelsForSpectra.clear();
   // This is the highest possible number of models used.
   const size_t allSources = datasets->numSourcesForSpectra();
   // To speed things along so we don't always have to check every detector
   // slot for every spectrum, first see which slots actually have active 
   // models associated with them, and store the number of sum comps for 
   // a given model so its lookup only needs to be performed once.
   std::vector<size_t> sourcesWithActiveMods;
   std::vector<size_t> numCompsForMods(allSources,0);
   for (size_t i=0; i<allSources; ++i) {
     string modName(models->lookupModelForSource(i+1));
     // Model with modName must be active/on or active/off.
     if (modName.length()) {
       const Model* mod = models->lookup(modName);
       if (mod && mod->isActive()) {
	 // Let's keep source numbers 1-based throughout.
	 sourcesWithActiveMods.push_back(i+1);
	 const size_t nComps = mod->sumComponents().size();
	 numCompsForMods[i] = nComps;
       }
     }
   }
   m_modelsForSpectra.resize(datasets->numberOfSpectra());
   // Loop over groups
   for (const auto& group : groupedSpectra) {
     std::set<size_t> sourcesForGroup;
     // Loop over spectra for this group
     for (const auto& spec : group) {
       const std::vector<const Model*> modsForSpec(models->getModsForSpec(spec));
       std::vector<const Model*>& spectraMods = m_modelsForSpectra[spec->spectrumNumber()-1];
       // modsForSpec has by definition size = allSources, and
       // may have many (or all) elements = 0.  We don't need to
       // check each one, thanks to sourcesWithActiveMods subset.
       // sourcesWithActiveMods is a vector of size_t
       for (const auto& actSource : sourcesWithActiveMods) {
	 const Model* mod = modsForSpec[actSource-1];
	 if (mod) {
	   spectraMods.push_back(mod);
	   // Don't want to repeat source nums, that's
	   // why this is a set and not a multiset.
	   sourcesForGroup.insert(mod->sourceNumber());
	 }
       }
       // end loop over spectra
     }

     // If more than one model, reserve 0 element for sum.
     size_t offset = sourcesForGroup.size() > 1 ? 1 : 0;
     std::map<size_t,size_t> assignments;
     size_t count = 0;
     // sourcesForGroup is a set of size_t
     for (const auto& num : sourcesForGroup) {
       assignments[num] = count + offset;
       ++count;
     }
     m_modelPositionDirectory.push_back(assignments);
     sourcesForGroup.clear();
     // end loop over groups
   }

}

PlotGroup* CreatePolarizationPlotGroups::initializePlotGroup (const size_t groupIndex,
							      std::vector<SpectralData*>& spectraForGroup)
{

   using namespace XSContainer;

   // This is guaranteed to return a non-null PlotGroup pointer, or
   // else it throws.

   // spectraForGroup contains all spectra that will make up plot group.
   // next check that we have equal numbers of I, Q, and U then calculate
   // the modulation factors which are the rmf/arf. assume that if multiple
   // sets of I,Q,U are included then the modulation factors are the same

   initEffectiveAreas(spectraForGroup);
   if ( (spectraForGroup.size() % 3) != 0 ) {
     std::ostringstream diag;
     diag << "The number of spectra in the plot group is " << spectraForGroup.size()
	  << " which is not divisible by 3.\n Use setplot group to make sure that"
	  << " the plot group contains sets of I,Q,U spectra.\n\n";
     throw YellowAlert(diag.str());
   }
   IntegerVector numStokes(3,0);
   RealArray mrfQ, arfI;

   for (const auto& spec : spectraForGroup) {

     if ( !spec->inxflt("Stokes") ) {
       std::ostringstream diag;
       diag << "Spectrum " << spec->spectrumNumber()
	    << " is missing an XFLTnnnn keyword specifying which Stokes parameter it is for.\n";
       throw YellowAlert(diag.str());
     }
     const size_t stokesPar = static_cast<size_t>(spec->xflt("Stokes"));
     if ( stokesPar < 0 || stokesPar > 2 ) {
       std::ostringstream diag;
       diag << "Spectrum " << spec->spectrumNumber()
	    << " has an invalid XFLTnnnn keyword specifying which Stokes parameter it is for.\n"
	    << " it should be Stokes:0, Stokes:1, or Stokes:2.\n";
       throw YellowAlert(diag.str());
     }
     numStokes[stokesPar]++;

     // if this is the I spectrum then store the arf contents in m_arf while if it is the
     // Q spectrum then store them in m_mrf
     if ( stokesPar == 0 ) {
       m_arf.resize(spec->channels());
       m_arf = spec->effectiveAreas();
     } else if ( stokesPar == 1 ) {
       m_mrf.resize(spec->channels());
       m_mrf = spec->effectiveAreas();
     }

   }

   PlotGroup* gr = 0;
   int Nchannels (0);
   size_t numModVects = m_modelPositionDirectory[groupIndex-1].size();
   if (spectraForGroup.empty())
   {
       gr = new PlotGroup(0, numModVects);
   }
   else
   {

       Nchannels = (*spectraForGroup.begin())->channels();
       // create a PlotGroup & allocate its arrays.
       gr = new PlotGroup(Nchannels,numModVects);
       // saveData will be needed as an internal work array during binning.
       gr->m_saveData.resize(Nchannels, 0.0);
       // Even if background is not to be displayed, background spectra
       // will still contribute to error calculation.  So always allocate
       // these vectors.
       gr->m_background.data.resize(Nchannels,0);
       gr->m_background.errors.push_back(std::vector<Real>(Nchannels,0));
   }

   gr->m_single = (Nchannels == 1);

   std::map<int,PlotSettings::RebinInfo>::const_iterator rbEntry = 
        m_settings->groupsRebinInfo().find(groupIndex); 
   if (rbEntry == m_settings->groupsRebinInfo().end())
   {
       // Use all-group default values, key = -1, always in map.
       rbEntry = m_settings->groupsRebinInfo().find(-1);
   }
   gr->m_critSignificance = rbEntry->second.sigma;
   gr->m_critNumChans = rbEntry->second.maxBins;
   gr->m_errorType    = rbEntry->second.mode; 

   int maxChan (0);
   int minChan (999999);

   for (const auto& spec : spectraForGroup)
   {
       // Find the first and last contributing channels.
       // For channel to contribute to a bin, it must 1) be noticed, 
       // and 2) if using setplot area, then area != 0.0.

       // We really shouldn't have to do this here.  SpectralData
       // shouldn't be waiting till prepareForFit to update its
       // m_indirectNotice member.
       std::valarray<size_t> indirectNotice;
       spec->buildIndirectNotice(indirectNotice);

       bool isFound = false;
       size_t j=0;
       while (!isFound && j<indirectNotice.size())
       {
          isFound = (spec->effectiveAreas()[indirectNotice[j]] != 0.0);
          if (!isFound)
             ++j;       
       }
       if (isFound)
       {
          minChan = std::min(minChan, static_cast<int>(indirectNotice[j]));
          // Now work backwards and find the max.
          // If we're here, we know indirectNotice isn't empty and that
          // we will find a contributing chan.  At worst it will be the
          // same as minChan.
          j  = indirectNotice.size()-1;
          isFound = false;
          while (!isFound)
          {
             isFound =  (spec->effectiveAreas()[indirectNotice[j]] != 0.0);
             if (!isFound)
                --j;
          } 
          maxChan = std::max(maxChan,static_cast<int>(indirectNotice[j]));
       }
   }

   // chanXSpacing is used to separate plot groups when they are
   // viewed in channel space.
   // lastInputChannel is 0-based.
   gr->m_channelPlotSpacing = (maxChan - minChan);
   if ( gr->m_channelPlotSpacing < 0 )
   {
       // group is totally ignored.
       gr->m_channelPlotSpacing = 10;
       gr->m_lastInputChannel = 0;
   }
   else if ( gr->m_single )
   {
       // single point (data contains observation through a filter,
       // not a spectrum).
       gr->m_channelPlotSpacing = 20;
       gr->m_lastInputChannel = 0;
   }
   else
   {
       // square up to nearest multiple of 10.
       gr->m_channelPlotSpacing = 10*( (gr->m_channelPlotSpacing+2)/10 + 1 );       
       gr->m_lastInputChannel = maxChan;
   }

   return gr;
}

bool CreatePolarizationPlotGroups::getNextPlotBin (PlotGroup* gr, int npts, const std::vector<SpectralData*>& groupSpectra, const std::map<size_t,size_t>& modAssignments, int& latestChannel)
{

  // This is called once for each plot bin.  Function returns TRUE if ANY noticed channels
  // have been added to the bin. A return value of FALSE implies that it has reached
  // the last channel before finding a bin contributor, and therefore no more calls
  // are required to this function for this particular PlotGroup.

  // This routine currently assumes that each plot group consists of a triplet of I,Q,U Stokes
  // parameter spectra

  // Also can ASSUME that gr has already been checked to contain AT LEAST 1 spectrum
  // with AT LEAST 1 noticed channel that is not nullified by the zero-area defect.

  // Input:  npts = the 0-based index of the current bin.
  //         latestChannel = the first 0-based channel to examine.
  // Output: latestChannel = 1 beyond the last examined channel.

  // gr->m_lastInputChannel = The highest 0-based noticed channel in any of the
  //        spectra within the group.  We can always assume that it is safe
  //        to access the i_th element from one of gr's vectors,
  //        where 0 <= i <= lastInputChannel.

  if (groupSpectra.empty() || latestChannel > gr->m_lastInputChannel) 
        return false;

  // if there is more than one source per spectrum, the models are stored
  // in the model std:vector<Real*> container enumerated from 1.
  // if there isn't then that vector only has one element, enumerated as zero.
  // for the latter case, the source number is 1 and the binned model is stored
  // in location zero. for the former case, the binned models are stores in
  // location == source number and the sum will ultimately be stored in location
  // zero.

  bool doneBin = false;
  bool notChannelSpace = m_settings->xOption() != CHANNELS;

  int firstChannelInBin = -1;
  int lastChannelInBin = -1;

  int chanEnd = gr->m_lastInputChannel;
  int numChannelsInBin = 0;

  // These will store the emin of the first channel in bin and
  // the emax of the last channel.  If energies are in descending order,
  // then startingEnergy > endingEnergy.
  Real startingEnergy = 0.0;
  Real endingEnergy = 0.0;

  Real critVar ( gr->m_critSignificance * gr->m_critSignificance );

  // local variables used in accumulation. for all the arrays index 0 is I, 1 is Q and 2 is U
  RealArray datapt(0.0,3);
  RealArray errorpt(0.0,3);

  Real effMu(0.0);

  // local variable to accumulate the models
  size_t nModels = gr->m_model.size();
  std::vector<RealArray> modelpt(3);
  for (size_t i=0; i<3; i++) {
    modelpt[i].resize(nModels);
    modelpt[i] = 0.0;
  }

  while (!doneBin ) {

    // Issue zero area warning at most one time for each spectrum.
    static std::vector<bool> issuedAreaWarning;
    if (latestChannel == 0) {
      //reset for new plot group
      issuedAreaWarning.clear();
      issuedAreaWarning.resize(groupSpectra.size(),false);
    }

    // loop over the spectra in this plot group
    for (const auto& current : groupSpectra) {
      size_t spectrumNumber (current->spectrumNumber());

      // find out whether this spectrum is I, Q, or U (ie xflt = 0,1, or 2)
      if ( !current->inxflt("Stokes") ) {
	std::ostringstream msg;
	msg <<"Spectrum "<<spectrumNumber<<" is missing an XFLTnnnn keyword.\n"
	    <<"   plotting polarization information needs this to specify which Stokes parameter the spectrum is for.\n";
	throw YellowAlert(msg.str());
      }
      const size_t stokesPar = static_cast<size_t>(current->xflt("Stokes"));

      const int offset = current->startChan() - current->firstChan();
      // if we got this far, detector must exist or be irrelevant.
      const Response* firstResp = 0;
      // This refers to the EBOUNDS energies.
      bool isEngDescending = false;
      std::vector<Response*>::const_iterator itResp = current->detectorResponseVector().begin();
      while (itResp != current->detectorResponseVector().end() && !firstResp) {
	// This could be a dummyrsp with no ebounds array.  Need to check.
	// But each spectrum in plot group must have at least one response
	// which passes this test if energy/wavelength is to be plotted.
	// Otherwise initializePlotGroup will have thrown.
	if (*itResp && (*itResp)->eboundsMin().size()) {
	  firstResp = *itResp;
	  isEngDescending = firstResp->eboundsMin()[0] >
	    firstResp->eboundsMin()[firstResp->eboundsMin().size()-1];
	}
	++itResp;
      }

      if ( current->noticedChannels(latestChannel + offset) ) {
	if (firstChannelInBin == -1) {
	  firstChannelInBin = latestChannel;
	  chanEnd = std::min(chanEnd,gr->m_critNumChans+latestChannel-1);
	  if (notChannelSpace) {
	    startingEnergy = isEngDescending ? firstResp->eboundsMax()[latestChannel] :
	      firstResp->eboundsMin()[latestChannel];
	  }
	}
	if ( notChannelSpace ) {
	  endingEnergy = isEngDescending ? firstResp->eboundsMin()[latestChannel] :
	    firstResp->eboundsMax()[latestChannel];
	}

	// accumulate data from SpectralData
	Real factor (current->exposureTime());
	if ( stokesPar == 0 ) {
	  factor /= m_arf[latestChannel];
	  effMu += (m_mrf[latestChannel]/m_arf[latestChannel])*current->spectrum(latestChannel)*factor;
	} else {
	  factor /= m_mrf[latestChannel];
	}
	Real specRate = current->spectrum(latestChannel);
	datapt[stokesPar] += factor*specRate;
	Real specVar = current->variance(latestChannel);
	if ( current->isSystematics() ) {
	  Real specSysfrac = current->fracsyserr(latestChannel);
	  specVar += specRate*specRate*specSysfrac*specSysfrac;
	}
	errorpt[stokesPar] += factor*factor*specVar;
	if ( current->background() ) {
	  const SpectralData* bckSpec = current->background()->data();
	  const RealArray& backgroundSpectrum = bckSpec->spectrum();
	  const RealArray& backgroundVariance = bckSpec->variance();
	  Real backContribution = backgroundSpectrum[latestChannel]*factor;
	  datapt[stokesPar] -=  backContribution;
	  Real backErrContribution = factor*factor*backgroundVariance[latestChannel];
	  errorpt[stokesPar] += backErrContribution;
	}
	if ( current->correction()) {
	  const RealArray& correctionSpectrum = current->correction()->spectrum();
	  datapt[stokesPar] -= current->correctionScale()*correctionSpectrum[latestChannel]*factor;
	}

	// loop over a vector of const Model pointers
	for (const auto& mod : m_modelsForSpectra[spectrumNumber-1]) {
	  const size_t sourceIndex = modAssignments.find(mod->sourceNumber())->second;
	  if (mod->isFoldedModelForNumber(spectrumNumber)) {
	    const RealArray& spectrumFit = mod->foldedModel(spectrumNumber).values;
	    modelpt[stokesPar][sourceIndex] += spectrumFit[latestChannel]*factor;
	  }
	}

	lastChannelInBin = latestChannel;
	++numChannelsInBin;

      } // end if noticed channel and no area defect

    } // end loop over spectra in group

    // test whether the current bin had finished accumulating - this is done based on the I Stokes parameter

    doneBin = ( latestChannel >= chanEnd ) ||
      ( datapt[0] > 0 && datapt[0]*datapt[0] >= errorpt[0]*critVar);

    ++latestChannel;
  } // end while !doneBin

  // Early exit to prevent possible divides by zero.
  if ( numChannelsInBin == 0 ) return false;

// Now set the min and max channel energies. This convoluted procedure is
// necessary to handle the cases of both channels increasing and decreasing
// with energy

  Real chanEmin = std::min(startingEnergy,endingEnergy);
  Real chanEmax = std::max(startingEnergy,endingEnergy);

  // ASSUME redshiftToSource value of -1.0 has been prevented.
  chanEmin *= (1.0 + m_settings->redshiftToSource());
  chanEmax *= (1.0 + m_settings->redshiftToSource());

  // calculate the polarization angle and fraction and store depending on which was requested

  effMu /= datapt[0];

  // fraction = sqrt(Q^2+U^2)/I
  Real polFrac(NODATA());
  if ( datapt[0] > 0.0 ) polFrac = sqrt(datapt[1]*datapt[1]+datapt[2]*datapt[2])/datapt[0];
  // angle (in radians) = 0.5 * arctan2(U,Q)
  Real polAngle = (180.0/M_PI) * 0.5 * atan2(datapt[2],datapt[1]);
  Real modulation = polFrac * effMu;
  Real polFracErr(NODATA()), polAngleErr(NODATA());
  if ( datapt[0] > 1.0 ) {
    // error on fraction = sqrt(W2 / I * (2. - modulation**2.) / ((I - 1.) * effective_mu**2))
    if ( (2.0-modulation*modulation) >= 0.0 ) {
      polFracErr = sqrt(errorpt[0]/datapt[0]
			* (2.0-modulation*modulation)/((datapt[0]-1.0)*effMu*effMu));
    }
    // error on angle = sqrt(0.5*W2/I/(I-1.0))/modulation
    polAngleErr = (180.0/M_PI) * sqrt(0.5*errorpt[0]/datapt[0]/(datapt[0]-1.0))/modulation;
  }
  if ( polFracErr == NODATA() ) polFrac = NODATA();
  if ( polAngleErr == NODATA() ) polAngle = NODATA();

  if ( m_isAngle ) {
    gr->m_yData.data[npts] = polAngle;
    gr->m_yData.errors[0][npts] = polAngleErr;
    for (size_t i=0; i<nModels; i++) {
      gr->m_model[i].data[npts] = (180.0/M_PI) * 0.5 * atan2(modelpt[2][i],modelpt[1][i]);
    }
  } else {
    gr->m_yData.data[npts] = polFrac;
    gr->m_yData.errors[0][npts] = polFracErr;
    for (size_t i=0; i<nModels; i++) {
      gr->m_model[i].data[npts] = sqrt(modelpt[1][i]*modelpt[1][i]+modelpt[2][i]*modelpt[2][i])/modelpt[0][i];
    }
  }

  // x axis and x bin width arrays. could be done earlier but better
  // to separate responsibilities.

  setXaxis(*m_settings,gr,firstChannelInBin,lastChannelInBin+1,chanEmin,chanEmax,npts);

  return true;
}

void CreatePolarizationPlotGroups::initEffectiveAreas (std::vector<SpectralData*>& spectraInGroup)
{
  for (auto& spectrum : spectraInGroup) spectrum->calcEffectiveAreas();
}

void CreatePolarizationPlotGroups::clearEffectiveAreas (std::vector<SpectralData*>& spectraInGroup)
{
  for (auto& spectrum : spectraInGroup) spectrum->clearEffectiveAreas();
}

std::vector<std::vector<SpectralData*> > CreatePolarizationPlotGroups::divideIntoGroups()
{
  std::vector<std::vector<SpectralData*> > spectraGroups;
  std::vector<SpectralData*> currentGroup;
  size_t currDG = -1;

  // loop round all spectra
  for (const auto& spec : XSContainer::datasets->allSpectralDataPointers()) {

    size_t specDG = spec->parent()->dataGroup();
    size_t specPG = spec->plotGroup();

    // if this is a new data group number then push the current group on to the vector
    // of groups and start a new group otherwise just add the spectrum to the current
    // group
    if ( specDG != currDG ) {
      if (!currentGroup.empty()) spectraGroups.push_back(currentGroup);
      currentGroup = {spec};
      currDG = specDG;
    } else {
      currentGroup.push_back(spec);
    }

    // check whether this spectrum has a data group or a plot group that matches
    // a spectrum that we have already placed in one of the spectraGroups. if so
    // add all the spectra in that group to the current group and delete them
    // from that group
    for (auto& group : spectraGroups) {
      if ( std::find_if(group.begin(), group.end(), [&](SpectralData*& other) {
	return (other->parent()->dataGroup() == specDG) || (other->plotGroup() == specPG);
      }) != group.end() ) {
	for (auto& other : group ) {
	  currentGroup.push_back(other);
	}
	group = {};
      }
    }

  }

  // if we are still accumulating a group then push that onto the vector of groups
  if ( !currentGroup.empty() ) spectraGroups.push_back(currentGroup);

  // remove any groups of spectra that have no members
  for (auto& group : spectraGroups) {
    if ( group.empty() ) {
      spectraGroups.erase(std::remove(spectraGroups.begin(), spectraGroups.end(), group), spectraGroups.end());
    }
  }

  return spectraGroups;
}



// Additional Declarations
