//   Read the documentation to learn more about C++ code generator
//   versioning.
//	  %X% %Q% %Z% %W%
#include <xsTypes.h>

// SpectralData
#include <XSModel/Data/SpectralData.h>
// Model
#include <XSModel/Model/Model.h>
// Weight
#include <Weight.h>
// Background
#include <XSModel/Data/BackCorr/Background.h>
// Model container
#include <XSModel/GlobalContainer/ModelContainer.h>
#include <XSContainer.h>
// Definition of SMALL
#include <XSsymbol.h>

#include <sstream>

Weight::~Weight()
{
  // body for abstract dtor. 
}

// Class Weight::ArrayInitializationError 

Weight::ArrayInitializationError::ArrayInitializationError (size_t num)
  : RedAlert(" initializing statistic workspace arrays:\n ")  
{
  std::ostringstream message;
  message << " Statistic calculation found spectrum number " << num << " array incorrect size ";
  reportAndExit(message.str(),-5);
}


// Class Weight 


Weight::Weight (const string& name)
  : m_name(name)
{
}

void Weight::calcDataVariance(const SpectralData* spectrum, RealArray& specVar) const
{
  // this method is a dummy since all the work is done by the version in the
  // inherited classes
}

void Weight::addModelVariance (const Model& model, const SpectralData* spectrum, 
			       RealArray& variance) const
{
  size_t spectrumNumber = spectrum->spectrumNumber();
  const std::valarray<size_t>& IN = spectrum->indirectNotice();
  
  // add model systematic error estimate to the variance.
  
  // spectrumNumber is used to find the correct dataset correspondence
  // in the Model object.
  Real msyst = XSContainer::models->modelSystematicError(model.name());
  if (model.isFoldedModelForNumber(spectrumNumber)) {
    if (variance.size() != IN.size()) {
      throw ArrayInitializationError(spectrumNumber);
    }        
    if (!model.foldedModel().empty() && model.foldedModel(spectrumNumber).variances.size()) {
      variance += model.foldedModel(spectrumNumber).variances[IN];
    }
    if (msyst > SMALL) {
      RealArray f (model.foldedModel(spectrumNumber).values[IN]);
      variance += msyst*msyst*f*f;       
    }
  }

}

void Weight::varianceDeriv (const Model& model, const SpectralData* spectrum, 
			    RealArray& dv) const
{
  const size_t spectrumNumber = spectrum->spectrumNumber();
  const std::valarray<size_t>& IN = spectrum->indirectNotice();
  size_t nbins = IN.size();
  if (dv.size() != nbins) dv.resize(nbins,0.);
  Real msyst = XSContainer::models->modelSystematicError(model.name());
  if ( msyst < SMALL) return;

  // model.detector() returns a map of size_t and pointers to Response objects
  for (const auto& ma : model.detector()) {
    size_t index = ma.first;
    if ( index == spectrumNumber ) {
      dv += msyst*msyst*model.foldedModel(index).values[IN];
    }
  }
}

// static function to create a new object
Weight* Weight::createWeightObject(const string& name)
{
  // this method creates and returns the pointer to the appropriate new 
  // object. note that if not enough characters are given in name for a unique
  // match then no object is created

  static string standard("standard");
  static string churazov("churazov");
  static string gehrels("gehrels");
  static string model("model");

  Weight* weightType = 0;
  if ( name == standard.substr(0,name.length()) ) { // standard
    weightType = new StandardWeight;
  } else if ( name == churazov.substr(0,name.length()) ) { // churazov
    weightType = new ChurazovWeight;
  } else if ( name == gehrels.substr(0,name.length()) ) { // gehrels
    weightType = new GehrelsWeight;
  } else if ( name == model.substr(0,name.length()) ) { // model
    weightType = new ModelWeight;
  } else {
    return 0;
  }

  return weightType;
}

// static function to check for a valid weight name. will also replace the
// input string with the full name of the weighting type.
bool Weight::isValidWeightName(string& name)
{
  bool valid(false);

  // try to create a Weight object
  try {
    Weight* weightType = createWeightObject(name);
    // if the created weight exists then input is valid
    if (weightType) {
      valid = true;
      name = weightType->name();
    }
    // tidy up
    delete weightType;
  } catch(...) {
    valid = false;
  }

  return valid;
}

// static function to write a list of the valid weight names
string Weight::weightNames ()
{
  return "standard | churazov | gehrels | model";
}



// Class StandardWeight 

StandardWeight::StandardWeight ()
  : Weight("standard")
{
}


StandardWeight::~StandardWeight()
{
}

void StandardWeight::calcDataVariance(const SpectralData* spectrum, RealArray& specVar) const
{
  // sum the variance from the spectrum and any background including any systematic
  // error on the spectrum
  const std::valarray<size_t>& IN = spectrum->indirectNotice();
  specVar = RealArray(spectrum->variance()[IN]);
  if ( spectrum->isSystematics() ) {
    RealArray systVar = spectrum->calcSystematicVariance();
    specVar += RealArray(systVar[IN]);
  }
  if ( spectrum->background() ) {
    const SpectralData* back = spectrum->background()->data();
    if ( back && back->variance().size() > 0 ) {
      specVar += back->variance()[IN];
    }
  }
}

void StandardWeight::addModelVariance (const Model& model, const SpectralData* spectrum, 
				       RealArray& variance) const
{
  // only need to add model error and systematics if required.      
  Weight::addModelVariance(model,spectrum,variance);
}

StandardWeight* StandardWeight::clone () const
{

  return new StandardWeight(*this);
}

// Class ModelWeight 

ModelWeight::ModelWeight ()
  : Weight("model")
{
}


ModelWeight::~ModelWeight()
{
}


void ModelWeight::calcDataVariance(const SpectralData* spectrum, RealArray& specVar) const
{
  // the source variance will be estimated from the model values so just need the
  // background variance here.
  const std::valarray<size_t>& IN = spectrum->indirectNotice();
  if ( spectrum->background() ) {
    const SpectralData* back = spectrum->background()->data();
    if ( back && back->variance().size() > 0 ) {
      specVar = back->variance()[IN];
    }
  }
}

void ModelWeight::addModelVariance (const Model& model, const SpectralData* spectrum,
				    RealArray& variance) const
{
  const size_t spectrumNumber = spectrum->spectrumNumber();
  const std::valarray<size_t>& IN = spectrum->indirectNotice();
  // add the model values to the variance as well as the observed variance on
  // the background spectrum
  if (model.isFoldedModelForNumber(spectrumNumber)) {
    if (variance.size() != IN.size()) {
      throw ArrayInitializationError(spectrumNumber);
    }
    Real T = spectrum->exposureTime();
    RealArray f = model.foldedModel(spectrumNumber).values[IN];

    variance += f / T;
    if (spectrum->background()) {
      const RealArray& backArray = spectrum->background()->data()->spectrum();
      variance += backArray / T;
    }
  }

  // add any additional model systematic error estimate to the variance.
  Weight::addModelVariance(model,spectrum,variance);
}

void ModelWeight::varianceDeriv (const Model& model, const SpectralData* spectrum, 
 				 RealArray& dv) const
{
  Weight::varianceDeriv(model,spectrum,dv);

  const size_t spectrumNumber = spectrum->spectrumNumber();
  const std::valarray<size_t>& IN = spectrum->indirectNotice();

  // model.detector() returns a map of size_t and pointers to Response objects
  for (const auto& ma : model.detector()) {
    size_t index = ma.first;
    if ( index == spectrumNumber ) {
      if ( dv.size() != IN.size()) dv.resize(IN.size());       
      dv += 1./(2.*spectrum->exposureTime());
    }
  }
}

ModelWeight* ModelWeight::clone () const
{
  
  return new ModelWeight(*this);
}

// Class GehrelsWeight 

GehrelsWeight::GehrelsWeight ()
  : Weight("gehrels")
{
}


GehrelsWeight::~GehrelsWeight()
{
}

void GehrelsWeight::calcDataVariance(const SpectralData* spectrum, RealArray& specVar) const
{
  static const Real OFFSET(0.75);
  const std::valarray<size_t>& IN = spectrum->indirectNotice();
  const RealArray& rate = spectrum->spectrum()[IN];
  const Real norm = spectrum->exposureTime();
  const size_t n = rate.size();
  // Doing these valarray arithmetic operations in a more tortured
  // manner to get around g++4.0.x runtime crashes with -O2 optimization.
  for (size_t i=0; i<n; ++i) {
    specVar[i] = (sqrt(norm*rate[i] + OFFSET) + 1.0)/norm;
    specVar[i] *= specVar[i];
  }
  // sum in any systematic error
  if ( spectrum->isSystematics() ) {
    const RealArray& systVar = spectrum->calcSystematicVariance();
    specVar += systVar[IN];
  }

  // sum in the background variance if required
  if ( spectrum->background() ) {
    const SpectralData* back = spectrum->background()->data();
    if ( back && back->variance().size() > 0 ) {
      const RealArray& backrate = back->spectrum()[IN];
      const RealArray& backnorm = back->exposureTime() 
	* (back->backgroundScale()[IN]/spectrum->backgroundScale()[IN])
	* (back->areaScale()[IN]/spectrum->areaScale()[IN]);
      for (size_t i=0; i<n; ++i) {
	Real backnorm_i = backnorm[i];
	Real backVar = (sqrt(backnorm_i*backrate[i] + OFFSET) + 1.0)/backnorm_i;
	backVar *= backVar;
	specVar[i] += backVar;
      }
    }
  }

}

void GehrelsWeight::addModelVariance (const Model& model, const SpectralData* spectrum,
				      RealArray& variance) const
{
  // only add any model error and systematics if required
  return Weight::addModelVariance(model,spectrum,variance);
}

GehrelsWeight* GehrelsWeight::clone () const
{

  return new GehrelsWeight(*this);
}

// Class ChurazovWeight 

ChurazovWeight::ChurazovWeight ()
  : Weight("churazov")
{
}


ChurazovWeight::~ChurazovWeight()
{
}

void ChurazovWeight::calcDataVariance(const SpectralData* spectrum, RealArray& specVar) const
{
  const std::valarray<size_t>& IN = spectrum->indirectNotice();
  const RealArray& rate = spectrum->spectrum()[IN];
  const Real norm = spectrum->exposureTime();
  const size_t n = rate.size();
  static const Real MAXCOUNTS = 20.0;
  for (size_t i = 0; i < n; ++i) {
    Real current = rate[i]*norm;
    bool terminate = current >= MAXCOUNTS;
    int lower = i;
    int upper = i;
    size_t bins  = 1;
    while (!terminate) {
      --lower;       
      ++upper;
      if (lower >=0 ) {
	++bins;
	current += rate[lower]*norm;      
      }
      if (upper < (int)n) {
	++bins;
	current += rate[upper]*norm;
      }
      terminate = ( (lower < 0 && upper >= (int)n) || current >= MAXCOUNTS);
    }
    specVar[i] = current / (bins*norm*norm);
  }
  // sum in any systematic error
  if ( spectrum->isSystematics() ) {
    const RealArray& systVar = spectrum->calcSystematicVariance();
    specVar += systVar[IN];
  }

  // sum in the background variance if required
  if ( spectrum->background() ) {
    const SpectralData* back = spectrum->background()->data();
    if ( back && back->variance().size() > 0 ) {
      const RealArray& backrate = back->spectrum()[IN];
      const RealArray& backnorm = back->exposureTime() 
	* (back->backgroundScale()[IN]/spectrum->backgroundScale()[IN])
	* (back->areaScale()[IN]/spectrum->areaScale()[IN]);
      for (size_t i = 0; i < n; ++i) {
	Real current = backrate[i]*backnorm[i];
	bool terminate = current >= MAXCOUNTS;
	int lower = i;
	int upper = i;
	size_t bins  = 1;
	while (!terminate) {
	  --lower;       
	  ++upper;
	  if (lower >=0 ) {
	    ++bins;
	    current += backrate[lower]*backnorm[lower];      
	  }
	  if (upper < (int)n) {
	    ++bins;
	    current += backrate[upper]*backnorm[upper];
	  }
	  terminate = ( (lower < 0 && upper >= (int)n) || current >= MAXCOUNTS);
	}
	Real backVar = current / (bins*backnorm[i]*backnorm[i]);
	specVar[i] += backVar;
      }
    }
  }

}

void ChurazovWeight::addModelVariance (const Model& model, const SpectralData* spectrum,
				       RealArray& variance) const
{
  // only add any model error and systematics if required
  return Weight::addModelVariance(model,spectrum,variance);
}

ChurazovWeight* ChurazovWeight::clone () const
{

  return new ChurazovWeight(*this);
}

