//
//  XSPEC12  November 2003
//
//

#include <XSFit/Fit/Fit.h>
#include <XSFit/Fit/FitMethod.h>
#include <XSFit/Fit/StatManager.h>
#include <XSModel/Data/DataSet.h>
#include <XSModel/Data/DataUtility.h>
#include <XSModel/Data/SpectralData.h>
#include <XSModel/Data/BackCorr/Background.h>
#include <XSModel/GlobalContainer/ModelContainer.h>
#include <XSModel/GlobalContainer/DataContainer.h>
#include <XSModel/GlobalContainer/ResponseContainer.h>
#include <XSModel/Model/Model.h>
#include <XSModel/Model/Component/Component.h>
#include <XSModel/Parameter/Parameter.h>
#include <XSModel/Parameter/ResponseParam.h>
#include <XSPlot/Plot/PlotDirector.h>
#include <XSStat/FitStatMethod.h>
#include <XSUser/Global/Global.h>        
#include <XSUser/Global/XSGlobal.h>        
#include <XSUser/UserInterface/xstcl.h>
#include <XSUser/Handler/XSinterface.h>
#include <XSUtil/Utils/XSstream.h>
#include <XSUtil/Utils/XSutility.h>
#include <XSUtil/Parse/XSparse.h>
#include <XSFunctions/Utilities/FunctionUtility.h>
#include <XSContainer.h>
#include <XSsymbol.h>
#include <XSstreams.h>
#include <xsTypes.h>
#include <XSUser/Handler/HandlerUtils.h>

#include <iomanip>
#include <ctime>
#include <cmath>
#include <cstring>

namespace {
  void showAbund();
  void showBayes();
  void showControl();
  void showFiles();
  void showFit();
  void showNoticed();
  void showParams(bool doFree, bool doFrozen, bool doLinked);
  void showPha();
  void showPlot();
  void showRates();
  void showRespPars(bool doFree, bool doFrozen, bool doLinked);
  void showResponse();
  void showSpecificParams(const StringArray& parIDs, bool doFree=true, bool doFrozen=true,
			  bool doLinked=true, bool isRespPar=false);
  void showXsect();
  void commonDataReport(const DataSet* ds, const SpectralData* sd);
  void rebinDisplay(int plotGroup, const PlotSettings::RebinInfo& rbInfo);
}

int
XSGlobal::xsShow(ClientData cdata,Tcl_Interp* tclInterp,int objc, Tcl_Obj* CONST  objv[])
{
   static string prevOption("all");

   StringArray rawArgs;
   HandlerUtils::tclArgsToCpp(objc, objv, rawArgs);

   doShow(rawArgs, prevOption);

   return TCL_OK;
}

// Declared in XSinterface.h
void XSGlobal::doShow(const StringArray& rawArgs, string& prevOption)
{
   using namespace std;
   using namespace XSContainer;

   string option;
   enum OPTIONS {ABUND, ALL, ALLFILE, BAYES, CONTROL, DATA, FILES, FIT, 
		 FREE, FROZEN, LINKED, MODEL, NOTICED, PARAMETER, PHA, 
		 PLOT, RATES, RESPONSE, RPARAMETER, TIED, VERSION, XSECT};
   static map<string, size_t> allowedOptions;
   allowedOptions["abund"] = ABUND;
   allowedOptions["all"] = ALL;
   allowedOptions["allfile"] = ALLFILE;
   allowedOptions["bayes"] = BAYES;
   allowedOptions["control"] = CONTROL;
   allowedOptions["data"] = DATA;
   allowedOptions["files"] = FILES;
   allowedOptions["fit"] = FIT;
   allowedOptions["free"] = FREE;
   allowedOptions["frozen"] = FROZEN;
   allowedOptions["linked"] = LINKED;
   allowedOptions["model"] = MODEL;
   allowedOptions["noticed"] = NOTICED;
   allowedOptions["parameters"] = PARAMETER;
   allowedOptions["pha"] = PHA;
   allowedOptions["plot"] = PLOT;
   allowedOptions["rates"] = RATES;
   allowedOptions["response"] = RESPONSE;
   allowedOptions["rparameters"] = RPARAMETER;
   allowedOptions["tied"] = TIED;
   allowedOptions["version"] = VERSION;
   allowedOptions["xsect"] = XSECT;

   const size_t nArgs = rawArgs.size();
   if (nArgs > 1) 
   {
           option = rawArgs[1]; 
           option = XSutility::lowerCase(option);
           if (option == "?")
           {
              XSutility::printValidOptions(tcout, string("show"),
                        allowedOptions);
              return;
           }
   }
   else option = prevOption;

   map<string,size_t>::const_iterator itOpt(allowedOptions.lower_bound(option));
   map<string,size_t>::const_iterator itOptEnd(allowedOptions.end());
   size_t selected = 0;
   if (itOpt != itOptEnd && itOpt->first.find(option) == 0) 
                selected = itOpt->second;
   else
   {
      tcout << "Unrecognized show option." << std::endl;
      XSutility::printValidOptions(tcout, string("show"), allowedOptions);
      return;
   }
   try
   {
      switch (selected)
      {
         case ABUND:
            showAbund();
            break;
         case ALL:
            doVersion(rawArgs);
            showControl();
            if (tpout.maxChatter() >= 15)
            {
               showAbund();
            }
            showPlot();
            showResponse();
            datasets->showData(true);
            models->showModel();
            showFit();
	    showBayes();
            break;
         case ALLFILE:
            showResponse();
            datasets->showData(true);
            break;
	 case BAYES:
	    showBayes();
	    break;
         case CONTROL:
            showControl();
            break;
         case DATA:
            datasets->showData(false);
            break;
         case FILES:
            showFiles();
            break;
         case FIT:
            showFit();
            break;
         case FREE:
            if (nArgs > 2) {
	      StringArray inArgs;
	      for (size_t i=2; i<nArgs; ++i) inArgs.push_back(rawArgs[i]);
	      IntegerVector dummy;
	      StringArray strPars;
	      XSparse::collectParams(inArgs,dummy,strPars);
	      showSpecificParams(strPars, true, false, false, false);
            } else
	      showParams(true, false, false);
            break;
         case FROZEN:
            if (nArgs > 2) {
	      StringArray inArgs;
	      for (size_t i=2; i<nArgs; ++i) inArgs.push_back(rawArgs[i]);
	      IntegerVector dummy;
	      StringArray strPars;
	      XSparse::collectParams(inArgs,dummy,strPars);
	      showSpecificParams(strPars, false, true, false, false);
            } else
	      showParams(false, true, false);
            break;
         case LINKED:
         case TIED:
            if (nArgs > 2) {
	      StringArray inArgs;
	      for (size_t i=2; i<nArgs; ++i) inArgs.push_back(rawArgs[i]);
	      IntegerVector dummy;
	      StringArray strPars;
	      XSparse::collectParams(inArgs,dummy,strPars);
	      showSpecificParams(strPars, false, false, true, false);
            } else
	      showParams(false, false, true);
            break;
         case MODEL:
         {
            // This code is mostly pasted from ModelContainer::showModel() and
            // particularly Model::printHeading() as part of a non-interface-changing
            // patch modification.  
            tcout << "\nCurrent model list:";
            if (models->modelSet().empty())
            {
               tcout << " none " << endl;
            }
            else
            {
               tcout << endl;
               map<string,bool>::const_iterator itModNames = 
                        models->activeModelNames().begin();
               map<string,bool>::const_iterator itNamesEnd = 
                        models->activeModelNames().end();
               while (itModNames != itNamesEnd)
               {
                  // mods size must always be >= 1 if name is in activeModelNames map.
                  vector<Model*> mods = models->lookupModelGroup(itModNames->first);
                  for (size_t iMod=0; iMod<mods.size(); ++iMod)
                  {
                     const Model* mod = mods[iMod];
                     if (iMod==0)
                     {
                        const string delim("()*+ \t");
                        const string& storedFullExpr = mod->fullExpression();
                        string reportFullExpr;
                        // Add bracketed component nums to full expression.
                        int compIdx = 0;
                        string::size_type startPos = 0;
                        do 
                        {
                           string::size_type wordPos = 
                                storedFullExpr.find_first_not_of(delim, startPos);
                           string::size_type endPos = string::npos;
                           if (wordPos != string::npos)
                           {
                              endPos = storedFullExpr.find_first_of(delim, wordPos);
                              string::size_type len = (endPos == string::npos) ?
                                        string::npos : endPos - startPos;
                              reportFullExpr += storedFullExpr.substr(startPos, len);
                              ++compIdx;
                              std::ostringstream oss;
                              oss << '<' << compIdx << '>';
                              reportFullExpr += oss.str(); 
                           }
                           else if (startPos != string::npos)
                           {
                              // We've reached the last component.  Pick up any
                              // trailing ')'s.
                              reportFullExpr += storedFullExpr.substr(startPos);
                           }
                           startPos = endPos;
                        }  while (startPos != string::npos); 

                        tcout << "\nModel ";
                        if (mod->name() != Model::DEFAULT() ) tcout << mod->name() << ':';
                        tcout << reportFullExpr;
                        if (datasets->numSourcesForSpectra() >= 1 ) 
                                tcout << " Source No.: " << mod->sourceNumber();
                        bool isAct = itModNames->second;
                        string status = isAct ? string("   Active") : string("   Inactive");
                        tcout << status << "/";
                        status = mod->isActive() ? string("On") : string("Off");
                        tcout << status << std::endl; 
                        tcout <<"      For Data Group(s): "; 
                     } // end if lowest group numbered model

                     tcout << mod->dataGroupNumber() <<" ";
                  }
                  tcout <<std::endl;
                  ++itModNames;
               }
            }
            tcout << std::endl;
            if (models->autonomousEnergy().size())
            {
               tcout << "   Using autonomous energy array.\n" << endl;
            }
            else if (models->isExtended())
            {
               tcout << "   Using extended response energies.\n" << endl;
            }
            else
            {
               tcout << "   Using energies from responses.\n" << endl;
            }
         }
            break;
         case NOTICED:
            showNoticed();
            break;
         case PARAMETER:
            if (nArgs > 2)
            {
               StringArray inArgs;
               for (size_t i=2; i<nArgs; ++i)
                  inArgs.push_back(rawArgs[i]);
               IntegerVector dummy;
               StringArray strPars;
               XSparse::collectParams(inArgs,dummy,strPars);
               showSpecificParams(strPars, true, true, true, false);
            } 
            else        
	       showParams(true, true, true);
            break;
         case PHA:
            showPha();
            break;
         case PLOT:
            showPlot();
            break;
         case RATES:
            showRates();
            break;
         case RESPONSE:
            showResponse();
            break;
         case RPARAMETER:
            if (nArgs > 2)
            {
               StringArray inArgs;
               for (size_t i=2; i<nArgs; ++i)
                  inArgs.push_back(rawArgs[i]);
               IntegerVector dummy;
               StringArray inPars;
               XSparse::collectParams(inArgs,dummy,inPars);
               showSpecificParams(inPars, true, true, true, true);
            } 
            else
	       showRespPars(true, true, true);
            break;
         case VERSION:
            // This is an allowed redundacy.  Just call the 'version'
            //   command handler.
            doVersion(rawArgs);
            break;
         case XSECT:
            showXsect();
            break;
         default:
            tcout << "Show option " << option << " does not exist" << endl;
            XSutility::printValidOptions(tcout, string("show"), allowedOptions);
            break;
      }
   }
   catch (YellowAlert&)
   {
   }
   if (itOpt != itOptEnd)
   {
      prevOption = itOpt->first;
   }
}

namespace {
   using namespace std;
   using namespace XSContainer;

   void showAbund()
   {
      tcout << "\nSolar Abundance Table: \n";
      const string& abund = FunctionUtility::ABUND();
      tcout << abund  << ": "  <<  FunctionUtility::abundDoc(abund) << '\n';

      const size_t& n = FunctionUtility::NELEMS();
      const vector<float>& abundances = FunctionUtility::abundanceVectors(abund);

      ios_base::fmtflags currentSetting(tcout.flags());
      streamsize p(tcout.precision(3));
      tcout.setf(ios_base::scientific | ios_base::showpoint | ios_base::uppercase );
      for ( size_t j = 0; j < n; ++j)
      {
              tcout << left << "  " << setw(4) 
                    <<  FunctionUtility::elements(j) + ":" <<
                    right << setw(10) << abundances[j];
              if ( j % 5 == 4 || j == n - 1) tcout << '\n';    
      }
      tcout << flush;
      tcout.flags(currentSetting);
      tcout.precision(p);
   }

   void showBayes()
   {
      const bool isOn = fit->statManager()->getBayes();
      if (isOn)
      {
         tcout << "\nBayes is on"<<std::endl;
	 tcout << "\n  Parameters with non-constant Bayes settings:"<<std::endl;
	 string allSettings;
	 fit->statManager()->printBayes(allSettings);
	 tcout << allSettings << std::endl;
      }
      else
      {
         tcout << "\nBayes is off\n"<<std::endl;
      }
   }
   
   void showControl()
   {
      time_t calndrTime;
      char timebuf[100];
      time(&calndrTime);
      strcpy(timebuf, ctime(&calndrTime));
      tcout << "\n" << timebuf;

      size_t saveFreq = XSGlobal::globalData->autoSaveFrequency();
      if (saveFreq == XSparse::NOTFOUND())
      {
         tcout << " Auto-saving is disabled." << endl;
      }
      else if (saveFreq == 1)
      {
         tcout << " Auto-saving is done after every command." << endl;
      }
      else
      {
         tcout << " Auto-saving is done after every " << saveFreq
                   << " commands." << endl;
      }

      if (fit->statManager()->usingSingleStat())
         tcout << " Fit statistic in use: " << 
	   fit->statManager()->usingSingleStat()->fullName() << endl;
      tcout << " Minimization technique: " << 
                   fit->fitMethod()->fullName() << endl;
      tcout << "    Convergence criterion = " << 
                           fit->fitMethod()->deltaCrit() << endl;
      Real parDelta = XSContainer::models->proportionalDelta();
      tcout << "    Parameter fit deltas: ";
      if (parDelta <= 0.0)
         tcout << "fixed values" << endl;
      else
         tcout << parDelta << " * parValue" << endl;
      if (fit->fitMethod()->delayedGratification())
	tcout << "    Using delayed gratification." << endl;

      if (!fit->useChainRule())
         tcout << "Not using chain rule to calculate statistic derivatives (slower method)." << std::endl;
      else
         tcout << "Using chain rule to calculate statistic derivatives (faster method)." << std::endl;

      if (fit->queryMode() == Fit::ON)
      {
         tcout << " Querying enabled." << endl;
      }
      else
      {
         if (fit->queryMode() == Fit::YES)
         {
            tcout << " Querying disabled - will continue fitting." << endl;
         }
         else
         {
            tcout << " Querying disabled - will not continue fitting." << endl;
         }
      }
      if (fit->renormType() == Fit::AUTO)
      {
         tcout << " Auto-renorming enabled." << endl;
      }
      else if (fit->renormType() == Fit::PREFIT)
      {
         tcout << " Prefit-renorming enabled." << endl;
      }           
      if (models->defaultModelSystematicError() != 0.0)
      {
         tcout << " Model default systematic error = " <<
                   models->defaultModelSystematicError() << endl;
      }

      tcout << " Solar abundance table: " << FunctionUtility::ABUND() << endl;
      showXsect();
      tcout << " Cosmology in use: H0 = " << models->cosmo().H0 
            << " q0 = " << models->cosmo().q0 << " Lambda0 = "
            << models->cosmo().lambda0 << endl
            << " Model data directory: " << 
                   FunctionUtility::modelDataPath() << endl;
   } //end showControl

   void showFiles()
   {
      // Essentially the same as showData, but outputs spectra in order
      // of DataSet file names rather than spectrum number.

       if (datasets->dataArray().empty())
       {
          tcout << "\n No Spectra defined." << endl;
       }
       else
       {
          string filx =  
                  datasets->dataArray().size() > 1 ? " files " : " file ";
          string 
                  spec =  datasets->numberOfSpectra() > 1 ? " spectra " : " spectrum ";
          tcout << '\n' << datasets->dataArray().size() << filx
                << datasets->numberOfSpectra() << spec << endl;

          DataArrayConstIt itDs = datasets->dataArray().begin();
          DataArrayConstIt itDsEnd = datasets->dataArray().end();
          while (itDs != itDsEnd)
          {
             if (itDs->second->isMultiple())
             {
                SpectralDataMapConstIt itSd = itDs->second->multiSpectralData().begin();
                SpectralDataMapConstIt itSdEnd = itDs->second->multiSpectralData().end();
                while (itSd != itSdEnd)
                {
                   tcout << endl;
                   itDs->second->report(itSd->first);
                   ++itSd;
                }

             }
             else
             {
                tcout << endl;
                itDs->second->report(0);
             }
             ++itDs;
          }
          tcout << flush;
       }
   }

   void showFit()
   {
      fit->statManager()->reportStats();

      if (!fit->useChainRule())
         tcout << " Fit is using the full (slower) differentiation method."
            << endl;
   }

   void showNoticed()
   {
       if (datasets->dataArray().empty())
       {
          tcout << "\n No Spectra defined." << endl;
       }
       else
       {
          const size_t nSpecs = datasets->numberOfSpectra();
          for (size_t i=1; i<=nSpecs; ++i)
          {
             const SpectralData* sd = datasets->lookup(i);
             tcout << endl << "Spectrum No. " << sd->spectrumNumber() << endl;
             commonDataReport(sd->parent(), sd);
             sd->reportNoticed();
          }
       }   
   }

   void showParams(bool doFree, bool doFrozen, bool doLinked)
   {
      bool saveFree = Parameter::printFree();
      bool saveFrozen = Parameter::printFrozen();
      bool saveLinked = Parameter::printLinked();
      Parameter::printFree(doFree);
      Parameter::printFrozen(doFrozen);
      Parameter::printLinked(doLinked);

      if (doFree && !doFrozen && !doLinked)
      {
         tcout << "\nFree parameters defined:";
      }
      else if (!doFree && doFrozen && !doLinked)
      {
         tcout << "\nFrozen parameters defined:";
      }
      else if (!doFree && !doFrozen && doLinked)
      {
         tcout << "\nLinked parameters defined:";
      }
      else
      {
         tcout << "\nParameters defined:"; 
      }
      const ModelMap& mods = models->modelSet();

      if (mods.empty())
      {
              tcout << " none" << endl;
      }
      else
      {
         map<string,bool>::const_iterator itModNames = models->activeModelNames().begin();
         map<string,bool>::const_iterator itNamesEnd = models->activeModelNames().end();
         while (itModNames != itNamesEnd)
         {
            vector<Model*> mods = models->lookupModelGroup(itModNames->first);
            // mods should always have size >=1 if in activeModelNames map.
            mods[0]->printHeading();
            for (size_t i=0; i<mods.size(); ++i)
            {
               tcout << *mods[i];
            }
            mods[0]->printMixComp();
            tcout << string(72,'_') << std::endl;
            ++itModNames;
         }  

         tcout << endl;     
      }

      if (responses->totalResponseParams())
      {
	showRespPars(doFree, doFrozen, doLinked);
      }

      Parameter::printFree(saveFree);
      Parameter::printFrozen(saveFrozen);
      Parameter::printLinked(saveLinked);

   } // end showParams


  void showRespPars(bool doFree, bool doFrozen, bool doLinked)
   {
      if (doFree && !doFrozen && !doLinked)
      {
         tcout << "\nFree response parameters defined:";
      }
      else if (!doFree && doFrozen && !doLinked)
      {
         tcout << "\nFrozen response parameters defined:";
      }
      else if (!doFree && !doFrozen && doLinked)
      {
         tcout << "\nLinked response parameters defined:";
      }
      else
      {
         tcout << "\nResponse parameters defined:";
      }

      if (!responses->totalResponseParams())
      {
         tcout << " none" << endl;
      }
      else
      {
         responses->reportResponseParams();
      }
   } // end showRespPars


   void showPha() 
   {
     if (datasets->dataArray().empty()) {
       tcout << "\n No Spectra defined." << endl;
       return;
     }

     const size_t nSpecs = datasets->numberOfSpectra();
     tcout << endl;
     ios_base::fmtflags saveFlags(tcout.flags());
     streamsize savePrec(tcout.precision());
     tcout << xsverbose(10);
     for (size_t i=1; i<=nSpecs; ++i) {
       const SpectralData* sd = datasets->lookup(i);
       tcout << " Information for spectrum " << i 
	     << "\n  belonging to plot group " << sd->plotGroup()
	     << ", data group " << sd->parent()->dataGroup() << endl;
       sd->reportKeywords();
       tcout << endl;
       tcout << setw(6) << "Bin " << setw(12) << "Source   " << setw(12) << "Error   ";
       if (sd->background())
	 tcout << setw(12) << "Back    " << setw(12) << "Error   ";
       if (sd->correction())
	 tcout << setw(12) << "Corr    ";
       tcout << setw(12) << "Net    " << setw(12) << "Error   ";
       bool isMod = false;
       // Don't assume sd->indirectNotice is up-to-date. It is only
       //  auto-updated if spectrum is active for fit.  Instead build
       // indirectNotice on-the-fly.
       std::valarray<size_t> indirectNotice;
       sd->buildIndirectNotice(indirectNotice);       
       const size_t szSpec = indirectNotice.size();
       RealArray foldedMod(.0,szSpec);
       if (XSContainer::models->modelSet().size()) {
	 for (size_t i=0; i<sd->detectorResponseVector().size(); ++i) {
	   const Response* resp = sd->detectorResponseVector()[i];
	   if (resp && resp->active()) {
	     const Model* mod = XSContainer::models->lookup(resp);
	     if (mod) {
               isMod = true;
               foldedMod += mod->foldedModel(sd->spectrumNumber()).values[indirectNotice];
	     }
	   }
	 }
       }
       if (isMod) tcout << setw(12) << "Model   ";
       tcout << endl;

       tcout << scientific << showpoint << setprecision(4);
       RealArray systVar = sd->calcSystematicVariance();
       bool isSyst = sd->isSystematics();
       for (size_t i=0; i<szSpec; ++i) {
	 const size_t iCh = indirectNotice[i];
	 Real phaObs = sd->spectrum(iCh);
	 Real phaVar = sd->variance(iCh);
	 if ( isSyst ) phaVar += systVar[iCh];
	 Real phaErr = sqrt(phaVar);
	 Real net = phaObs;
	 Real netErr = phaErr;
	 tcout << setw(6) << iCh+1 << setw(12) << phaObs << setw(12) << phaErr;

	 if (sd->background()) {
	   Real phaBack = sd->background()->data()->spectrum(iCh);
	   Real phaBackErr =sqrt(sd->background()->data()->variance(iCh));
	   net -= phaBack;
	   netErr = sqrt(sd->variance(iCh) + sd->background()->data()->variance(iCh));
	   tcout << setw(12) << phaBack << setw(12) << phaBackErr;
	 }
	 if (sd->correction()) {
	   Real phaCorr = sd->correctionScale()*sd->correction()->data()->spectrum(iCh);
	   net -= phaCorr;
	   tcout << setw(12) << phaCorr;
	 }
	 tcout << setw(12) << net << setw(12) << netErr;
	 if (isMod) {
	   // This array only contains the noticed channels, so use 
	   // index i and not iCh.
	   tcout << setw(12) << foldedMod[i];
	 }
	 tcout <<endl;
       }
       tcout << endl;
     }
     tcout << xsverbose();
     tcout.precision(savePrec);
     tcout.flags(saveFlags);
   }

   void showPlot()
   {
      using XSContainer::plot;
      const PlotSettings& settings = plot->setplot();
      tcout << " Plot settings:" << endl;
      string onOrOff = settings.getShowAddComponent() ? string("ON."):string("OFF.");
      tcout << "   Showing of individual additive components is " << onOrOff << endl;
      onOrOff = settings.getShowBackground() ? string("ON.") : string("OFF.");
      tcout << "   Showing of background spectra is " << onOrOff << endl;
      onOrOff = settings.getDivideByArea() ? string("ON."):string("OFF.");
      tcout << "   Effective area normalization is " << onOrOff << endl;
      tcout << "   Current unit settings:" << endl;
      tcout << "      Energy     = " << settings.getUnitID(ENERGY) << endl;
      tcout << "      Wavelength = " << settings.getUnitID(WAVELENGTH)
            << ", with Y-Axis displayed per ";
      if (settings.isWavePerHz())
         tcout << "Hz" << endl;
      else
         tcout << "length" << endl;
      string xMode;
      if (settings.xOption() == ENERGY)
         xMode = "Energy";
      else if (settings.xOption() == WAVELENGTH)
         xMode = "Wavelength";
      else
         xMode = "Channels";
      tcout << "   X-Axis data display mode: " << xMode << endl;
      tcout << "   Spectra plots will be shifted to source frame by redshift value z: "
                << settings.redshiftToSource() << endl;
      tcout << "   Device: " << plot->getPlottingDeviceName() << endl;
      onOrOff = settings.getShowLineIDs() ? string("ON."):string("OFF.");
      tcout << "   Plotting of line IDs is " << onOrOff << endl;
      if (settings.getShowLineIDs())
      {
         tcout << "     Temperature: " << settings.temperature()  
               << "     Emissivity Limit: " << settings.emisLimit() 
               << "     Redshift: " << settings.redshiftLinesToObs()
               << "     Low Energy: " << settings.IDLowEnergy()
               << "     High Energy: " << settings.IDHighEnergy()
	       << endl;
      }
      onOrOff = settings.splashPage() ? string("ON."):string("OFF.");
      tcout << "   Splashpage is " << onOrOff << endl;
      onOrOff = settings.xLog() ? string("ON."):string("OFF.");
      tcout << "   xlog for data plots is " << onOrOff << endl;
      onOrOff = settings.yLog() ? string("ON."):string("OFF.");
      tcout << "   ylog for data plots is " << onOrOff << endl;

      tcout <<"\n   Default plot rebin settings for all plot groups:"<<endl;      
      map<int,PlotSettings::RebinInfo>::const_iterator itRb = 
                settings.groupsRebinInfo().find(-1);
      // Entry for -1 should ALWAYS be in plot rebin map.
      tcout << "   Min. Signif.   Max. # Bins   Error Type"<< endl; 
      rebinDisplay(-1, itRb->second);
      map<int,PlotSettings::RebinInfo>::const_iterator itRbEnd = 
                settings.groupsRebinInfo().end();
      ++itRb;
      if (itRb != itRbEnd)
      {      
         tcout << "   With overridden rebin settings for the following plot groups:"
                 << endl;
         tcout << "   Groups    Min. Signif.   Max. # Bins    Error Type"<< endl;
         while (itRb != itRbEnd)
         {
            rebinDisplay(itRb->first, itRb->second);
            ++itRb;
         }
      }
   }  // end showPlot

   void showRates()
   {
       if (datasets->dataArray().empty())
       {
          tcout << "\n No Spectra defined." << endl;
       }
       else
       {
          const size_t nSpecs = datasets->numberOfSpectra();
          for (size_t i=1; i<=nSpecs; ++i)
          {
             const SpectralData* sd = datasets->lookup(i);
             tcout << endl;
             commonDataReport(sd->parent(), sd);
             sd->reportRates(); 
             tcout << " Spectral data counts: " << sd->totalFlux()*sd->exposureTime()
                   << endl;
             models->reportModelRate(sd);
          }
       } 
   }

   void showResponse()
   {
      // primitive data printing option
      tcout << "\n Responses read: ";
      if (responses->responseList().size() == 1)
      {
         tcout << " none " << endl;
         if (responses->responseList().begin()->first != DUMMY_RSP)
         {
            throw RedAlert(" Dummy Response Corrupted or deleted ");
         }
      }
      else
      {
         tcout << '\n';
         ResponseMapConstIter resp = responses->responseList().begin();
         while (resp != responses->responseList().end())
         {
            if ( resp->first != DUMMY_RSP)
            {
               Response* current(resp->second);
               string respName(resp->first);
               if (respName == USR_DUMMY_RSP)
                  respName = "Dummy response";
               tcout << "   " << respName << " associated with spectrum "
                     << current->spectrumNumber() << " source " 
                     << current->sourceNumber();
               if (current->dataGroup() > 1)
               {
                       tcout << " data group " <<  current->dataGroup();      
               }
               tcout << "\n      energies: " << current->numEnergies() 
                     << " channels: " << current->numChannels() << endl;
               if (current->isGainApplied())
               {
                  tcout << "   With applied gain:  slope = " 
                     << current->gainFactor()[1] << "  offset = " 
                     << current->gainFactor()[0] << endl;
               }
               if (current->getConstGain() || current->getLinearGain())
               {
                  tcout << "   Gain parameters will be adjusted by fit." << endl;
               }
            }
            ++resp;                       
         }

         RMFMapConstIter rmf = responses->RMFmap().begin();
         tcout << " Distinct RMF files: \n";
         while (rmf != responses->RMFmap().end())
         {
            tcout << "     " << rmf->first ;
            ++rmf;                       
         }

         tcout << endl;
      }
   }  // end showResponse

   void showSpecificParams(const StringArray& parIDs, bool doFree, bool doFrozen,
			   bool doLinked, bool isRespPar)
   {
     // set up which parameters are printable and save the current state so we
     // can restore it at the end of the routine
     bool saveFree = Parameter::printFree();
     bool saveFrozen = Parameter::printFrozen();
     bool saveLinked = Parameter::printLinked();
     Parameter::printFree(doFree);
     Parameter::printFrozen(doFrozen);
     Parameter::printLinked(doLinked);

     if (doFree && !doFrozen && !doLinked) {
         tcout << "\nFree parameters defined:";
      } else if (!doFree && doFrozen && !doLinked) {
         tcout << "\nFrozen parameters defined:";
      } else if (!doFree && !doFrozen && doLinked) {
         tcout << "\nLinked parameters defined:";
      } else {
         tcout << "\nParameters defined:"; 
      }
      string prevModName;
      const size_t linelen(72);
      bool doComp(false);
      bool doGroup(false);
      bool first(true);
      for (size_t i=0; i<parIDs.size(); ++i) {

	if ( parIDs[i] == "comp" ) {
	  doComp = true;
	} else if ( parIDs[i] == "group" ) {
	  doGroup = true;
	} else {
	  // if we get to here then we are just reading numbers
	  // which may include a model name. if doComp is set
	  // they are component numbers and if doGroup is set
	  // they are group numbers
	  string modName;
	  IntegerVector range;
	  XSparse::stringRangePair(parIDs[i],modName,range);
	  // If user enters "show par alpha:1,2", assume "alpha" also
	  // applies to "2".
	  const bool isNewName = first || (modName.length() && modName != prevModName);
	  if (modName.length())
            prevModName = modName;
	  else
            modName = prevModName;
	  if (isNewName) {
            // Place bottom bar under previous output
            if ( !first )
	      tcout << string(linelen,'_') << std::endl;
            if (isRespPar) {
	      // modName string must be a source number in this context.
	      size_t sourceNum = 1;
	      if (modName.length())
		sourceNum = XSutility::isInteger(modName);
	      if (sourceNum == string::npos || sourceNum == 0) {
		string errMsg("Invalid source number specifier: ");
		errMsg += modName + '\n';
		throw YellowAlert(errMsg);
	      }
	      if (sourceNum <= XSContainer::responses->respParContainer().size())
		tcout << string(linelen,'=') << "\nSource No.: " << sourceNum
                      << "\nRpar Spectrum Rmodel   Rpar_name  Unit   Value\n" 
                      << std::endl;
	      else {
		string errMsg("Out of range source number specifier: ");
		errMsg += modName + '\n';
		throw YellowAlert(errMsg);
	      }
	      // end if resppar
	    } else {
	      // Just get any model object from the group given by modName.
	      // It doesn't matter which for printing the heading.
	      const string searchName = modName.length() ? modName : Model::DEFAULT();
	      const Model* mod = XSContainer::models->lookup(searchName);
	      if (mod)
		mod->printHeading();
	      else {
		string err = modName.length() ?  "Cannot find model: " + modName + "\n"
		  : string("No unnamed models loaded.  Must enter <mod name>:<par num>\n");
		throw YellowAlert(err);
	      }
            }
	  } // end if new name

	  // if doComp and doGroup are not set then range is the set of parameters we
	  // need. if one of them is set then we need to find the actual range of parameters

	  IntegerVector rangePar;
	  if ( doComp ) {
	    // loop round components accumulating the parameters
	    for (size_t j=0; j<range.size(); j++) {
	      size_t iComp = static_cast<size_t>(range[j]);
	      const Component* specificComp = HandlerUtils::lookupStringIntObj<Component>(modName,iComp);
	      if ( specificComp ) {
		// we need a pointer to the model
		const Model* specificMod = HandlerUtils::getModFromName(modName);
		// and its source number
		const size_t sourceNum = specificMod->sourceNumber();
	        // and the total number of parameters for this model
		size_t nTotPar = specificMod->numberOfParameters();
		// pnums contains the parameter numbers for the component for the lowest datagroup
		IntegerVector pnums = specificComp->parameterIndices();
		// now iterate through the datagroups
		std::map<size_t,std::map<size_t,size_t> >::const_iterator
		  itSource = datasets->sourceToDgs().find(sourceNum);
		std::map<size_t,size_t>::const_iterator itGroup = itSource->second.begin();
		std::map<size_t,size_t>::const_iterator itGroupEnd = itSource->second.end();
		while (itGroup != itGroupEnd) {
		  size_t groupPos = itGroup->second;
		  size_t offset = (groupPos-1)*nTotPar;
		  for (size_t ip=0; ip<pnums.size(); ip++) rangePar.push_back(pnums[ip]+offset);
		  itGroup++;
		}
	      }
	    }
	  } else if ( doGroup ) {
	    // accumulate the parameters for the requested groups
	    const Model* specificMod = HandlerUtils::getModFromName(modName);
	    const size_t sourceNum = specificMod->sourceNumber();
	    // now iterate through the datagroups for this model
	    std::map<size_t,std::map<size_t,size_t> >::const_iterator
	      itSource = datasets->sourceToDgs().find(sourceNum);
	    std::map<size_t,size_t>::const_iterator itGroup = itSource->second.begin();
	    std::map<size_t,size_t>::const_iterator itGroupEnd = itSource->second.end();
	    while (itGroup != itGroupEnd) {
	      size_t groupNum = itGroup->first;
	      size_t groupPos = itGroup->second;
	      bool match = false;
	      for (size_t j=0; j<range.size(); j++) {
		if ( range[j] == (int)groupNum ) match = true;
	      }
	      // if there is a datagroup match then add all the parameters in this model
	      // for this datagroup
	      if ( match ) {
		size_t nTotPar = specificMod->numberOfParameters();
		size_t firstPar = specificMod->parameterIndexBase() + 1;
		firstPar += (groupPos-1)*nTotPar;
		for (size_t ip=0; ip<nTotPar; ip++) rangePar.push_back(ip+firstPar);
	      }
	      itGroup++;
	    }
	  } else {
	    rangePar.resize(range.size());
	    rangePar = range;
	  }

	  // We will want the total number of data groups
	  size_t numDataGroups = datasets->numberOfGroups();
	  size_t curDataGroup(0);
	  // Loop round the selected parameters
	  for (size_t j=0; j<rangePar.size(); ++j) {
            if (rangePar[j] < 1) {
	      // Can get here if user entered wildcards, which
	      // this isn't handling.
	      std::ostringstream oss;
	      oss << "Invalid parameter index: " << rangePar[j] << "\n";
	      throw YellowAlert(oss.str());
            }
            size_t iPar = static_cast<size_t>(rangePar[j]);
            const Parameter* specificPar = isRespPar ?
	      HandlerUtils::lookupStringIntObj<ResponseParam>(modName,iPar)
	      : HandlerUtils::lookupStringIntObj<Parameter>(modName,iPar);
            if (specificPar) {
	      size_t dataGroup = specificPar->dataGroup();
	      if ( (numDataGroups > 1) && !isRespPar && ( (j==0) || (curDataGroup != dataGroup) ) )
		tcout << "                           Data group: " << dataGroup << std::endl;
	      curDataGroup = dataGroup;
	      if ( specificPar->printable() ) tcout <<*specificPar;
            }
	  }
	  if (i == parIDs.size()-1)
            tcout << string(linelen,'_') << '\n' << std::endl;
	  first = false;
	}
      } // end args loop

      // restore which parameters are printable to the state at the start of the routine
      Parameter::printFree(saveFree);
      Parameter::printFrozen(saveFrozen);
      Parameter::printLinked(saveLinked);

   } // end showSpecificParams

   void showXsect()
   {
      tcout << " Photoionization Cross-Section Table:\n    ";
      tcout << FunctionUtility::XSECT() << ": " 
                 <<  FunctionUtility::crossSections(FunctionUtility::XSECT()) << endl;    
   }

   void commonDataReport(const DataSet* ds, const SpectralData* sd)
   {
      tcout << "Spectral Data File: " << ds->dataName();
      if (sd->rowNumber() > 0 ) tcout << "{"  << sd->rowNumber()  <<"}";
      tcout << "\n Assigned to Data Group " << ds->dataGroup()
           << " and Plot Group " << sd->plotGroup() << endl;
   }   

   void rebinDisplay(int plotGroup, const PlotSettings::RebinInfo& rbInfo)
   {
      string modeStr;
      switch (rbInfo.mode)
      {
         case PlotSettings::ROOTN:
            modeStr = "sqrt";
            break;
         case PlotSettings::GEHRELS1:
            modeStr = "poiss-1";
            break;
         case PlotSettings::GEHRELS2:
            modeStr = "poiss-2";
            break;
         case PlotSettings::GEHRELSM:
            modeStr = "poiss-3";
            break;
         default:
            modeStr = "quad";
            break;                
      }
      ios_base::fmtflags saveFmt(tcout.flags());
      if (plotGroup != -1)
         tcout << setw(9) << plotGroup;
      tcout << showpoint << setw(15) <<rbInfo.sigma<<setw(14)<<rbInfo.maxBins 
                <<setw(13)<< modeStr << endl;  
      tcout.flags(saveFmt); 
   }
}
